@@ -17,7 +17,6 @@
 
 from __future__ import absolute_import, division, print_function
 
-import math
 from collections import defaultdict
 from enum import IntEnum
 
@@ -65,6 +64,21 @@ class Status(IntEnum):
         return "Status.%s" % (self.name,)
 
 
+@attr.s(frozen=True, slots=True)
+class StructuralCoverageTag(object):
+    label = attr.ib()
+
+
+STRUCTURAL_COVERAGE_CACHE = {}
+
+
+def structural_coverage(label):
+    try:
+        return STRUCTURAL_COVERAGE_CACHE[label]
+    except KeyError:
+        return STRUCTURAL_COVERAGE_CACHE.setdefault(label, StructuralCoverageTag(label))
+
+
 class Example(object):
     """Examples track the hierarchical structure of draws from the byte stream,
     within a single test run.
@@ -655,6 +669,9 @@ class _Overrun(object):
     def __repr__(self):
         return "Overrun"
 
+    def as_result(self):
+        return self
+
 
 Overrun = _Overrun()
 
@@ -703,6 +720,7 @@ class ConjectureResult(object):
     extra_information = attr.ib()
     has_discards = attr.ib()
     target_observations = attr.ib()
+    tags = attr.ib()
     forced_indices = attr.ib(repr=False)
     examples = attr.ib(repr=False)
 
@@ -712,6 +730,9 @@ class ConjectureResult(object):
         self.index = len(self.buffer)
         self.forced_indices = frozenset(self.forced_indices)
 
+    def as_result(self):
+        return self
+
 
 # Masks for masking off the first byte of an n-bit buffer.
 # The appropriate mask is stored at position n % 8.
@@ -723,10 +744,10 @@ class ConjectureData(object):
     @classmethod
     def for_buffer(self, buffer, observer=None):
         return ConjectureData(
-            prefix=buffer, max_length=len(buffer), parameter=None, observer=observer,
+            prefix=buffer, max_length=len(buffer), random=None, observer=observer,
         )
 
-    def __init__(self, max_length, prefix, parameter, observer=None):
+    def __init__(self, max_length, prefix, random, observer=None):
         if observer is None:
             observer = DataObserver()
         assert isinstance(observer, DataObserver)
@@ -738,9 +759,9 @@ class ConjectureData(object):
         self.__block_starts = defaultdict(list)
         self.__block_starts_calculated_to = 0
         self.__prefix = prefix
-        self.__parameter = parameter
+        self.__random = random
 
-        assert parameter is not None or max_length <= len(prefix)
+        assert random is not None or max_length <= len(prefix)
 
         self.blocks = Blocks(self)
         self.buffer = bytearray()
@@ -766,6 +787,11 @@ class ConjectureData(object):
         # ConjectureRunner.generate_new_examples and fed to TargetSelector.
         self.target_observations = {}
 
+        # Tags which indicate something about which part of the search space
+        # this example is in. These are used to guide generation.
+        self.tags = set()
+        self.labels_for_structure_stack = []
+
         # Normally unpopulated but we need this in the niche case
         # that self.as_result() is Overrun but we still want the
         # examples for reporting purposes.
@@ -807,6 +833,7 @@ class ConjectureData(object):
                 else None,
                 has_discards=self.has_discards,
                 target_observations=self.target_observations,
+                tags=frozenset(self.tags),
                 forced_indices=self.forced_indices,
             )
             self.blocks.transfer_ownership(self.__result)
@@ -880,6 +907,7 @@ class ConjectureData(object):
             self.max_depth = self.depth
         self.__example_record.start_example(label)
         self.consecutive_discard_counts.append(0)
+        self.labels_for_structure_stack.append({label})
 
     def stop_example(self, discard=False):
         if self.frozen:
@@ -890,6 +918,15 @@ class ConjectureData(object):
         self.depth -= 1
         assert self.depth >= -1
         self.__example_record.stop_example(discard)
+
+        labels_for_structure = self.labels_for_structure_stack.pop()
+
+        if not discard:
+            if self.labels_for_structure_stack:
+                self.labels_for_structure_stack[-1].update(labels_for_structure)
+            else:
+                self.tags.update([structural_coverage(l) for l in labels_for_structure])
+
         if self.consecutive_discard_counts:
             # We block long sequences of discards. This helps us avoid performance
             # problems where there is rejection sampling. In particular tests which
@@ -982,9 +1019,9 @@ class ConjectureData(object):
             index = self.__bytes_drawn
             buf = self.__prefix[index : index + n_bytes]
             if len(buf) < n_bytes:
-                buf += uniform(self.__parameter.random, n_bytes - len(buf))
+                buf += uniform(self.__random, n_bytes - len(buf))
         else:
-            buf = self.__parameter.draw_bytes(n_bytes)
+            buf = uniform(self.__random, n_bytes)
         buf = bytearray(buf)
         self.__bytes_drawn += n_bytes
 
@@ -1052,87 +1089,3 @@ def bits_to_bytes(n):
     Equivalent to (n + 7) // 8, but slightly faster. This really is
     called enough times that that matters."""
     return (n + 7) >> 3
-
-
-generation_parameters_count = 0
-
-
-class GenerationParameters(object):
-    """Parameters to control generation of examples."""
-
-    AVERAGE_ALPHABET_SIZE = 3
-
-    ALPHABET_FACTOR = math.log(1.0 - 1.0 / AVERAGE_ALPHABET_SIZE)
-
-    def __init__(self, random):
-        self.random = random
-        self.__pure_chance = None
-        self.__alphabet = {}
-
-        global generation_parameters_count
-        generation_parameters_count += 1
-
-        self.__id = generation_parameters_count
-
-    def __repr__(self):
-        return "GenerationParameters(%d)" % (self.__id,)
-
-    def draw_bytes(self, n):
-        """Draw an n-byte block from the distribution defined by this instance
-        of generation parameters."""
-        alphabet = self.alphabet(n)
-
-        if alphabet is None:
-            return self.__draw_without_alphabet(n)
-
-        return self.random.choice(alphabet)
-
-    def __draw_without_alphabet(self, n):
-        return uniform(self.random, n)
-
-    def alphabet(self, n_bytes):
-        """Returns an alphabet - a list of values to use for all blocks with
-        this number of bytes - or None if this value should be generated
-        without an alphabet.
-
-        This is designed to promote duplication in the test case that would
-        otherwise happen with very low probability.
-        """
-        try:
-            return self.__alphabet[n_bytes]
-        except KeyError:
-            pass
-
-        if self.random.random() <= self.pure_chance:
-            # Sometiems we don't want to use an alphabet (e.g. for generating
-            # sets of integers having a small alphabet is disastrous), so with
-            # some probability we want to generate choices that do not use the
-            # alphabet. As with other factors we set this probability globally
-            # across the whole choice of distribution so we have various levels
-            # of mixing.
-            result = None
-        else:
-            # We draw the size as a geometric distribution with average size
-            # GenerationParameters.AVERAGE_ALPHABET_SIZE.
-            size = (
-                int(
-                    math.log(self.random.random())
-                    / GenerationParameters.ALPHABET_FACTOR
-                )
-                + 1
-            )
-            assert size > 0
-
-            size = self.random.randint(1, 10)
-            result = [self.__draw_without_alphabet(n_bytes) for _ in hrange(size)]
-
-        self.__alphabet[n_bytes] = result
-        return result
-
-    @property
-    def pure_chance(self):
-        """Returns a probability with which any given draw_bytes call should
-        be forced to be all pure."""
-        if self.__pure_chance is None:
-            self.__pure_chance = self.random.random()
-        return self.__pure_chance
