@@ -31,7 +31,7 @@ from hypothesis.internal.compat import ceil, hbytes, int_from_bytes
 from hypothesis.internal.conjecture.data import (
     ConjectureData,
     ConjectureResult,
-    GenerationParameters,
+    DataObserver,
     Overrun,
     Status,
     StopTest,
@@ -42,6 +42,7 @@ from hypothesis.internal.conjecture.datatree import (
     TreeRecordingObserver,
 )
 from hypothesis.internal.conjecture.junkdrawer import clamp
+from hypothesis.internal.conjecture.pareto import NO_SCORE, ParetoFront
 from hypothesis.internal.conjecture.shrinker import Shrinker, sort_key
 from hypothesis.internal.healthcheck import fail_health_check
 from hypothesis.reporting import base_report
@@ -51,9 +52,6 @@ from hypothesis.reporting import base_report
 __tracebackhide__ = True
 
 
-NO_SCORE = float("-inf")
-
-
 MAX_SHRINKS = 500
 CACHE_SIZE = 10000
 MUTATION_POOL_SIZE = 100
@@ -110,6 +108,17 @@ class ConjectureRunner(object):
         self.tree = DataTree()
 
         self.best_observed_targets = defaultdict(lambda: NO_SCORE)
+        self.best_examples_of_observed_targets = {}
+
+        # We keep the pareto front in the example database if we have one. This
+        # is only marginally useful at present, but speeds up local development
+        # because it means that large targets will be quickly surfaced in your
+        # testing.
+        if self.database_key is not None and self.settings.database is not None:
+            self.pareto_front = ParetoFront(self.random)
+            self.pareto_front.on_evict(self.on_pareto_evict)
+        else:
+            self.pareto_front = None
 
         # We want to be able to get the ConjectureData object that results
         # from running a buffer without recalculating, especially during
@@ -117,6 +126,10 @@ class ConjectureRunner(object):
         # executed test case.
         self.__data_cache = LRUReusedCache(CACHE_SIZE)
 
+    @property
+    def should_optimise(self):
+        return Phase.target in self.settings.phases
+
     def __tree_is_exhausted(self):
         return self.tree.is_exhausted
 
@@ -159,12 +172,30 @@ class ConjectureRunner(object):
 
         self.debug_data(data)
 
+        if self.pareto_front is not None and self.pareto_front.add(data.as_result()):
+            self.save_buffer(data.buffer, sub_key=b"pareto")
+
         assert len(data.buffer) <= BUFFER_SIZE
 
         if data.status >= Status.VALID:
             for k, v in data.target_observations.items():
                 self.best_observed_targets[k] = max(self.best_observed_targets[k], v)
 
+                if k not in self.best_examples_of_observed_targets:
+                    self.best_examples_of_observed_targets[k] = data.as_result()
+                    continue
+
+                existing_example = self.best_examples_of_observed_targets[k]
+                existing_score = existing_example.target_observations[k]
+
+                if v < existing_score:
+                    continue
+
+                if v > existing_score or sort_key(data.buffer) < sort_key(
+                    existing_example.buffer
+                ):
+                    self.best_examples_of_observed_targets[k] = data.as_result()
+
         if data.status == Status.VALID:
             self.valid_examples += 1
 
@@ -214,6 +245,9 @@ class ConjectureRunner(object):
 
         self.record_for_health_check(data)
 
+    def on_pareto_evict(self, data):
+        self.settings.database.delete(self.pareto_key, data.buffer)
+
     def generate_novel_prefix(self):
         """Uses the tree to proactively generate a starting sequence of bytes
         that we haven't explored yet for this test.
@@ -306,9 +340,9 @@ class ConjectureRunner(object):
                 HealthCheck.too_slow,
             )
 
-    def save_buffer(self, buffer):
+    def save_buffer(self, buffer, sub_key=None):
         if self.settings.database is not None:
-            key = self.database_key
+            key = self.sub_key(sub_key)
             if key is None:
                 return
             self.settings.database.save(key, hbytes(buffer))
@@ -317,13 +351,20 @@ class ConjectureRunner(object):
         if self.settings.database is not None and self.database_key is not None:
             self.settings.database.move(self.database_key, self.secondary_key, buffer)
 
-    @property
-    def secondary_key(self):
-        return b".".join((self.database_key, b"secondary"))
+    def sub_key(self, sub_key):
+        if self.database_key is None:
+            return None
+        if sub_key is None:
+            return self.database_key
+        return b".".join((self.database_key, sub_key))
 
     @property
-    def covering_key(self):
-        return b".".join((self.database_key, b"coverage"))
+    def secondary_key(self):
+        return self.sub_key(b"secondary")
+
+    @property
+    def pareto_key(self):
+        return self.sub_key(b"pareto")
 
     def note_details(self, data):
         self.__data_cache[data.buffer] = data.as_result()
@@ -429,29 +470,39 @@ class ConjectureRunner(object):
             )
             desired_size = max(2, ceil(0.1 * self.settings.max_examples))
 
-            for extra_key in [self.secondary_key, self.covering_key]:
-                if len(corpus) < desired_size:
-                    extra_corpus = list(self.settings.database.fetch(extra_key))
+            if len(corpus) < desired_size:
+                extra_corpus = list(self.settings.database.fetch(self.secondary_key))
 
-                    shortfall = desired_size - len(corpus)
+                shortfall = desired_size - len(corpus)
 
-                    if len(extra_corpus) <= shortfall:
-                        extra = extra_corpus
-                    else:
-                        extra = self.random.sample(extra_corpus, shortfall)
-                    extra.sort(key=sort_key)
-                    corpus.extend(extra)
+                if len(extra_corpus) <= shortfall:
+                    extra = extra_corpus
+                else:
+                    extra = self.random.sample(extra_corpus, shortfall)
+                extra.sort(key=sort_key)
+                corpus.extend(extra)
 
             for existing in corpus:
-                last_data = ConjectureData.for_buffer(
-                    existing, observer=self.tree.new_observer()
-                )
-                try:
-                    self.test_function(last_data)
-                finally:
-                    if last_data.status != Status.INTERESTING:
-                        self.settings.database.delete(self.database_key, existing)
-                        self.settings.database.delete(self.secondary_key, existing)
+                data = self.cached_test_function(existing)
+                if data.status != Status.INTERESTING:
+                    self.settings.database.delete(self.database_key, existing)
+                    self.settings.database.delete(self.secondary_key, existing)
+
+            # If we've not found any interesting examples so far we try some of
+            # the pareto front from the last run.
+            if len(corpus) < desired_size and not self.interesting_examples:
+                desired_extra = desired_size - len(corpus)
+                pareto_corpus = list(self.settings.database.fetch(self.pareto_key))
+                if len(pareto_corpus) > desired_extra:
+                    pareto_corpus = self.random.sample(pareto_corpus, desired_extra)
+                pareto_corpus.sort(key=sort_key)
+
+                for existing in pareto_corpus:
+                    data = self.cached_test_function(existing)
+                    if data not in self.pareto_front:
+                        self.settings.database.delete(self.pareto_key, existing)
+                    if data.status == Status.INTERESTING:
+                        break
 
     def exit_with(self, reason):
         self.debug("exit_with(%s)" % (reason.name,))
@@ -467,12 +518,12 @@ class ConjectureRunner(object):
             # the time to look for additional failures.
             return
 
+        self.debug("Generating new examples")
+
         zero_data = self.cached_test_function(hbytes(BUFFER_SIZE))
         if zero_data.status > Status.OVERRUN:
             self.__data_cache.pin(zero_data.buffer)
 
-        self.optimise_all(zero_data)
-
         if zero_data.status == Status.OVERRUN or (
             zero_data.status == Status.VALID and len(zero_data.buffer) * 2 > BUFFER_SIZE
         ):
@@ -494,14 +545,6 @@ class ConjectureRunner(object):
         self.health_check_state = HealthCheckState()
 
         def should_generate_more():
-            # If we haven't found a bug, keep looking.  We check this before
-            # doing anything else as it's by far the most common case.
-            if not self.interesting_examples:
-                return True
-            # If we've found a bug and won't report more than one, stop looking.
-            elif not self.settings.report_multiple_bugs:
-                return False
-            assert self.first_bug_found_at <= self.last_bug_found_at <= self.call_count
             # End the generation phase where we would have ended it if no bugs had
             # been found.  This reproduces the exit logic in `self.test_function`,
             # but with the important distinction that this clause will move on to
@@ -510,8 +553,23 @@ class ConjectureRunner(object):
             if (
                 self.valid_examples >= self.settings.max_examples
                 or self.call_count >= max(self.settings.max_examples * 10, 1000)
+                or (
+                    self.best_examples_of_observed_targets
+                    and self.valid_examples * 2 >= self.settings.max_examples
+                    and self.should_optimise
+                )
             ):  # pragma: no cover
                 return False
+
+            # If we haven't found a bug, keep looking - if we hit any limits on
+            # the number of tests to run that will raise an exception and stop
+            # the run.
+            if not self.interesting_examples:
+                return True
+            # If we've found a bug and won't report more than one, stop looking.
+            elif not self.settings.report_multiple_bugs:
+                return False
+            assert self.first_bug_found_at <= self.last_bug_found_at <= self.call_count
             # Otherwise, keep searching for between ten and 'a heuristic' calls.
             # We cap 'calls after first bug' so errors are reported reasonably
             # soon even for tests that are allowed to run for a very long time,
@@ -520,24 +578,6 @@ class ConjectureRunner(object):
                 self.first_bug_found_at + 1000, self.last_bug_found_at * 2
             )
 
-        # GenerationParameters are a set of decisions we make that are global
-        # to the whole test case, used to bias the data generation in various
-        # ways. This is an approach very very loosely inspired by the paper
-        # "Swarm testing." by Groce et al. in that it induces deliberate
-        # correlation between otherwise independent decisions made during the
-        # generation process.
-        #
-        # More importantly the generation is designed to make certain scenarios
-        # more likely (e.g. small examples, duplicated values), which can help
-        # or hurt in terms of finding interesting things. Whenever the result
-        # of our generation is a bad test case, for whatever definition of
-        # "bad" we like (currently, invalid or too large), we ditch the
-        # parameter early. This allows us to potentially generate good test
-        # cases significantly more often than we otherwise would, by selecting
-        # for parameters that make them more likely.
-        parameter = GenerationParameters(self.random)
-        count = 0
-
         # We attempt to use the size of the minimal generated test case starting
         # from a given novel prefix as a guideline to generate smaller test
         # cases for an initial period, by restriscting ourselves to test cases
@@ -608,7 +648,7 @@ class ConjectureRunner(object):
                 # starting from the new novel prefix that has discovered.
                 try:
                     trial_data = self.new_conjecture_data(
-                        prefix=prefix, parameter=parameter, max_length=max_length
+                        prefix=prefix, max_length=max_length
                     )
                     self.tree.simulate_test_function(trial_data)
                     continue
@@ -629,42 +669,125 @@ class ConjectureRunner(object):
             else:
                 max_length = BUFFER_SIZE
 
-            data = self.new_conjecture_data(
-                prefix=prefix, parameter=parameter, max_length=max_length
-            )
+            data = self.new_conjecture_data(prefix=prefix, max_length=max_length)
 
             self.test_function(data)
 
-            self.optimise_all(data)
-
-            count += 1
+            # A thing that is often useful but rarely happens by accident is
+            # to generate the same value at multiple different points in the
+            # test case.
+            #
+            # Rather than make this the responsibility of individual strategies
+            # we implement a small mutator that just takes parts of the test
+            # case with the same label and tries replacing one of them with a
+            # copy of the other and tries running it. If we've made a good
+            # guess about what to put where, this will run a similar generated
+            # test case with more duplication.
             if (
-                data.status < Status.VALID
-                or len(data.buffer) * 2 >= BUFFER_SIZE
-                or count > 5
+                # An OVERRUN doesn't have enough information about the test
+                # case to mutate, so we just skip those.
+                data.status >= Status.INVALID
+                # This has a tendency to trigger some weird edge cases during
+                # generation so we don't let it run until we're done with the
+                # health checks.
+                and self.health_check_state is None
             ):
-                count = 0
-                parameter = GenerationParameters(self.random)
+                initial_calls = self.call_count
+                failed_mutations = 0
+                while (
+                    should_generate_more()
+                    # We implement fairly conservative checks for how long we
+                    # we should run mutation for, as it's generally not obvious
+                    # how helpful it is for any given test case.
+                    and self.call_count <= initial_calls + 5
+                    and failed_mutations <= 5
+                ):
+                    groups = defaultdict(list)
+                    for ex in data.examples:
+                        groups[ex.label, ex.depth].append(ex)
 
-    def optimise_all(self, data):
-        """If the data or result object is suitable for hill climbing, run hill
-        climbing on all of its target observations."""
-        if data.status == Status.VALID:
-            for target in data.target_observations:
+                    groups = [v for v in groups.values() if len(v) > 1]
+
+                    if not groups:
+                        break
+
+                    group = self.random.choice(groups)
+
+                    ex1, ex2 = sorted(
+                        self.random.sample(group, 2), key=lambda i: i.index
+                    )
+                    assert ex1.end <= ex2.start
+
+                    replacements = [data.buffer[e.start : e.end] for e in [ex1, ex2]]
+
+                    replacement = self.random.choice(replacements)
+
+                    try:
+                        # We attempt to replace both the the examples with
+                        # whichever choice we made. Note that this might end
+                        # up messing up and getting the example boundaries
+                        # wrong - labels matching are only a best guess as to
+                        # whether the two are equivalent - but it doesn't
+                        # really matter. It may not achieve the desired result
+                        # but it's still a perfectly acceptable choice sequence.
+                        # to try.
+                        new_data = self.cached_test_function(
+                            data.buffer[: ex1.start]
+                            + replacement
+                            + data.buffer[ex1.end : ex2.start]
+                            + replacement
+                            + data.buffer[ex2.end :],
+                            # We set error_on_discard so that we don't end up
+                            # entering parts of the tree we consider redundant
+                            # and not worth exploring.
+                            error_on_discard=True,
+                            extend=BUFFER_SIZE,
+                        )
+                    except ContainsDiscard:
+                        failed_mutations += 1
+                        continue
+
+                    if (
+                        new_data.status >= data.status
+                        and data.buffer != new_data.buffer
+                        and all(
+                            k in new_data.target_observations
+                            and new_data.target_observations[k] >= v
+                            for k, v in data.target_observations.items()
+                        )
+                    ):
+                        data = new_data
+                        failed_mutations = 0
+                    else:
+                        failed_mutations += 1
+
+    def optimise_targets(self):
+        """If any target observations have been made, attempt to optimise them
+        all."""
+        if not self.should_optimise:
+            return
+        while True:
+            prev_calls = self.call_count
+            for target, data in list(self.best_examples_of_observed_targets.items()):
                 self.new_optimiser(data, target).run()
+            # This shows up as a missing branch for some reason but is definitely
+            # covered if you add an assertion to check.
+            if self.interesting_examples or (prev_calls == self.call_count):
+                break
 
     def _run(self):
         self.reuse_existing_examples()
         self.generate_new_examples()
+        self.optimise_targets()
         self.shrink_interesting_examples()
         self.exit_with(ExitReason.finished)
 
-    def new_conjecture_data(self, prefix, parameter, max_length=BUFFER_SIZE):
+    def new_conjecture_data(self, prefix, max_length=BUFFER_SIZE, observer=None):
         return ConjectureData(
             prefix=prefix,
-            parameter=parameter,
             max_length=max_length,
-            observer=self.tree.new_observer(),
+            random=self.random,
+            observer=observer or self.tree.new_observer(),
         )
 
     def new_conjecture_data_for_buffer(self, buffer):
@@ -758,15 +881,24 @@ class ConjectureRunner(object):
 
         return Optimiser(self, example, target)
 
-    def cached_test_function(self, buffer):
+    def cached_test_function(self, buffer, error_on_discard=False, extend=0):
         """Checks the tree to see if we've tested this buffer, and returns the
         previous result if we have.
 
         Otherwise we call through to ``test_function``, and return a
         fresh result.
+
+        If ``error_on_discard`` is set to True this will raise ``ContainsDiscard``
+        in preference to running the actual test function. This is to allow us
+        to skip test cases we expect to be redundant in some cases. Note that
+        it may be the case that we don't raise ``ContainsDiscard`` even if the
+        result has discards if we cannot determine from previous runs whether
+        it will have a discard.
         """
         buffer = hbytes(buffer)[:BUFFER_SIZE]
 
+        max_length = min(BUFFER_SIZE, len(buffer) + extend)
+
         def check_result(result):
             assert result is Overrun or (
                 isinstance(result, ConjectureResult) and result.status != Status.OVERRUN
@@ -778,16 +910,34 @@ class ConjectureRunner(object):
         except KeyError:
             pass
 
-        rewritten, status = self.tree.rewrite(buffer)
+        if error_on_discard:
+
+            class DiscardObserver(DataObserver):
+                def kill_branch(self):
+                    raise ContainsDiscard()
+
+            observer = DiscardObserver()
+        else:
+            observer = DataObserver()
+
+        dummy_data = self.new_conjecture_data(
+            prefix=buffer, max_length=max_length, observer=observer
+        )
 
         try:
-            result = check_result(self.__data_cache[rewritten])
-        except KeyError:
+            self.tree.simulate_test_function(dummy_data)
+        except PreviouslyUnseenBehaviour:
             pass
         else:
-            assert result.status != Status.OVERRUN or result is Overrun
-            self.__data_cache[buffer] = result
-            return result
+            if dummy_data.status > Status.OVERRUN:
+                dummy_data.freeze()
+                try:
+                    return self.__data_cache[dummy_data.buffer]
+                except KeyError:
+                    pass
+            else:
+                self.__data_cache[buffer] = Overrun
+                return Overrun
 
         # We didn't find a match in the tree, so we need to run the test
         # function normally. Note that test_function will automatically
@@ -795,17 +945,11 @@ class ConjectureRunner(object):
 
         result = None
 
-        if status != Status.OVERRUN:
-            data = self.new_conjecture_data_for_buffer(buffer)
-            self.test_function(data)
-            result = check_result(data.as_result())
-            assert status is None or result.status == status, (status, result.status)
-            status = result.status
-        if status == Status.OVERRUN:
-            result = Overrun
-
-        assert result is not None
-
+        data = self.new_conjecture_data(
+            prefix=max((buffer, dummy_data.buffer), key=len), max_length=max_length,
+        )
+        self.test_function(data)
+        result = check_result(data.as_result())
         self.__data_cache[buffer] = result
         return result
 
@@ -819,3 +963,7 @@ class ConjectureRunner(object):
         result = str(event)
         self.events_to_strings[event] = result
         return result
+
+
+class ContainsDiscard(Exception):
+    pass
